#!/usr/bin/env python3

import asyncio
import json
import sys
from argparse import ArgumentParser, Namespace
from datetime import datetime
from pathlib import Path, PurePath

from reolinkfw import __version__, firmware_info, firmwares_from_file

HW_FIELDS = ("board_type", "detail_machine_type", "board_name")


async def info(args: Namespace) -> None:
    pak_infos = await firmware_info(args.file_or_url, not args.no_cache)
    if args.json is None:
        width = 21
        for idx, info in enumerate(pak_infos):
            if "error" in info:
                raise Exception(info["error"])
            info = Namespace(**info)
            fs_types = set(fs["type"] for fs in info.filesystems)
            fs_names = [fs["name"] for fs in info.filesystems]
            version = f"{info.firmware_version_prefix}.{info.version_file}"
            hw_names = set(getattr(info, key) for key in HW_FIELDS)
            build_date = datetime.strptime(info.build_date, "%y%m%d").date()
            print(info.pak)
            print(f"{'Model:':{width}}", info.display_type_info)
            print(f"{'Hardware info:':{width}}", ', '.join(sorted(hw_names)))
            print(f"{'Device type:':{width}}", info.type)
            print(f"{'Firmware version:':{width}}", version)
            print(f"{'Build date:':{width}}", build_date)
            print(f"{'Architecture:':{width}}", info.architecture)
            print(f"{'OS:':{width}}", info.os)
            print(f"{'Kernel image name:':{width}}", info.kernel_image_name)
            print(f"{'Linux banner:':{width}}", info.linux_banner)
            print(f"{'U-Boot version:':{width}}", info.uboot_version or "Unknown")
            print(f"{'U-Boot compiler:':{width}}", info.uboot_compiler or "Unknown")
            print(f"{'U-Boot linker:':{width}}", info.uboot_linker or "Unknown")
            print(f"{'File system:':{width}}", ', '.join(sorted(fs_types)))
            print(f"{'File system sections:':{width}}", ', '.join(fs_names))
            print(f"{'Board vendor:':{width}}", info.board_vendor or "Unknown")
            print(f"{'Board:':{width}}", info.board or "Unknown")
            if idx != len(pak_infos) - 1:
                print()
    else:
        indent = None if args.json < 0 else args.json
        print(json.dumps(pak_infos, indent=indent, default=str))


async def extract(args: Namespace) -> None:
    fws = await firmwares_from_file(args.file_or_url, not args.no_cache)
    if not fws:
        raise Exception("No PAKs found in ZIP file")
    dest = Path.cwd() if args.dest is None else args.dest
    for pakname, fw in fws:
        with fw:
            name = fw.sha256() if pakname is None else PurePath(pakname).stem
            await asyncio.to_thread(fw.extract, dest / name, args.force)


def main() -> None:
    parser = ArgumentParser(description="Extract information and files from Reolink firmwares")
    parser.add_argument("-V", "--version", action="version", version=f"%(prog)s {__version__}")
    subparsers = parser.add_subparsers(dest="command", required=True)

    pcache = ArgumentParser(add_help=False)
    pcache.add_argument("--no-cache", action="store_true", help="don't use cache for remote files (URLs)")

    parser_i = subparsers.add_parser("info", parents=[pcache], aliases=['i'])
    parser_i.add_argument("file_or_url", help="URL or on-disk file")
    parser_i.add_argument("-j", "--json", nargs='?', type=int, const=-1, metavar="indent", help="JSON output with optional indentation level for pretty print")
    parser_i.set_defaults(func=info)

    descex = "Extract the file system and a few other files from a Reolink firmware"
    parser_e = subparsers.add_parser("extract", parents=[pcache], aliases=['e'], help=descex.lower(), description=descex)
    parser_e.add_argument("file_or_url", help="URL or on-disk file")
    parser_e.add_argument("-d", "--dest", type=Path, help="destination directory. Default: current directory")
    parser_e.add_argument("-f", "--force", action="store_true", help="overwrite existing files. Does not apply to UBIFS. Default: %(default)s")
    parser_e.set_defaults(func=extract)

    args = parser.parse_args()
    try:
        asyncio.run(args.func(args))
    except Exception as e:
        sys.exit(f"error: {e}")


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Mapping
from pathlib import PurePosixPath
from stat import filemode
from typing import Any, Iterator, Literal, Optional

from ubireader.ubifs import ubifs, walk
from ubireader.ubifs.defines import (
    UBIFS_COMMON_HDR_SZ,
    UBIFS_ITYPE_BLK,
    UBIFS_ITYPE_CHR,
    UBIFS_ITYPE_DIR,
    UBIFS_ITYPE_FIFO,
    UBIFS_ITYPE_LNK,
    UBIFS_ITYPE_REG,
    UBIFS_ITYPE_SOCK,
    UBIFS_NODE_MAGIC,
    UBIFS_ROOT_INO,
    UBIFS_SB_NODE,
    UBIFS_SB_NODE_SZ,
)
from ubireader.ubifs.nodes import common_hdr, ino_node, sb_node
from ubireader.ubifs.output import _process_reg_file
from ubireader.ubi_io import ubi_file
from ubireader.utils import guess_leb_size

from reolinkfw.tmpfile import TempFile
from reolinkfw.typedefs import Buffer, FileDescriptorOrPath, StrPath


class File:

    def __init__(self, image: UBIFS, nodes: Mapping[str, Any], name: str = '', parent: Optional[Directory] = None) -> None:
        self._image = image
        self._nodes = nodes
        self._name = name
        self._parent = parent

    @property
    def name(self) -> str:
        return self._name

    @property
    def parent(self) -> Optional[Directory]:
        return self._parent

    @property
    def path(self) -> PurePosixPath:
        if self._parent is None:
            return PurePosixPath('/')
        return self._parent.path / self._name

    @property
    def inode(self) -> ino_node:
        return self._nodes["ino"]

    @property
    def mode(self) -> int:
        return self.inode.mode

    @property
    def filemode(self) -> str:
        return filemode(self.mode)

    @property
    def is_dir(self) -> bool:
        return False

    @property
    def is_file(self) -> bool:
        return False

    @property
    def is_symlink(self) -> bool:
        return False


class DataFile(File):

    def read_bytes(self) -> bytes:
        ...

    def read_text(self, encoding: str = "utf8", errors: str = "strict") -> str:
        return self.read_bytes().decode(encoding, errors)


class RegularFile(DataFile):

    @property
    def is_file(self) -> Literal[True]:
        return True

    def read_bytes(self) -> bytes:
        return _process_reg_file(self._image._ubifs, self._nodes, None)


class Symlink(DataFile):

    @property
    def is_symlink(self) -> Literal[True]:
        return True

    def read_bytes(self) -> bytes:
        return self.inode.data

    def readlink(self) -> PurePosixPath:
        return PurePosixPath(self.read_text())


class Directory(File):

    def __init__(self, image: UBIFS, nodes: Mapping[str, Any], name: str = '', parent: Optional[Directory] = None) -> None:
        super().__init__(image, nodes, name, parent)
        self._children: dict[str, File] = {}
        for dent in nodes.get("dent", []):
            cls = filetype[dent.type]
            self._children[dent.name] = cls(image, image.inodes[dent.inum], dent.name, self)

    def __iter__(self) -> Iterator[File]:
        yield from self._children.values()

    @property
    def is_dir(self) -> Literal[True]:
        return True

    @property
    def children(self) -> dict[str, File]:
        return self._children

    def select(self, path: StrPath) -> Optional[File]:
        """Select a file of any kind by path.

        The path can be absolute or relative.
        Special entries `'.'` and `'..'` are supported.
        """
        path = PurePosixPath(path)
        if str(path) == "..":
            return self.parent if self.parent is not None else self
        if path.root == '/':
            if str(self.path) == '/':
                path = path.relative_to('/')
            else:
                return self._image.root.select(path)
        if str(path) == '.':
            return self
        child, *descendants = path.parts
        if (file := self._children.get(child)) is not None:
            if isinstance(file, Directory) and descendants:
                return file.select(PurePosixPath(*descendants))
            elif not descendants:
                return file
        return None

    def riter(self) -> Iterator[File]:
        yield self
        for file in self:
            if isinstance(file, Directory):
                yield from file.riter()
            else:
                yield file


class UBIFS:

    def __init__(self, ubifs: ubifs) -> None:
        self._ubifs = ubifs
        self._inodes: dict[int, dict[str, Any]] = {}
        self._bad_blocks = []
        walk.index(ubifs, ubifs.master_node.root_lnum, ubifs.master_node.root_offs, self._inodes, self._bad_blocks)
        self._root = Directory(self, self._inodes[UBIFS_ROOT_INO])

    def __enter__(self) -> UBIFS:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def __iter__(self) -> Iterator[File]:
        yield from self._root.riter()

    @property
    def inodes(self) -> dict[int, dict[str, Any]]:
        return self._inodes

    @property
    def root(self) -> Directory:
        return self._root

    def close(self) -> None:
        self._ubifs._file._fhandle.close()

    def select(self, path: StrPath) -> Optional[File]:
        return self._root.select(path)

    @classmethod
    def from_file(cls, path: FileDescriptorOrPath) -> UBIFS:
        return cls(ubifs(ubi_file(path, guess_leb_size(path))))

    @classmethod
    def from_bytes(cls, bytes_: Buffer, offset: int = 0) -> UBIFS:
        chdr = common_hdr(bytes_[offset:offset+UBIFS_COMMON_HDR_SZ])
        if chdr.magic != int.from_bytes(UBIFS_NODE_MAGIC, "little") or chdr.node_type != UBIFS_SB_NODE:
            raise Exception("Not UBIFS")
        sb_start = offset + UBIFS_COMMON_HDR_SZ
        sb_end = sb_start + UBIFS_SB_NODE_SZ
        sblk = sb_node(bytes_[sb_start:sb_end])
        tmpfile = TempFile(bytes_[offset:])
        tmpfile.open()  # tmpfile's close() will not be called.
        return cls(ubifs(ubi_file(tmpfile, sblk.leb_size)))


filetype = {
    UBIFS_ITYPE_REG: RegularFile,
    UBIFS_ITYPE_DIR: Directory,
    UBIFS_ITYPE_LNK: Symlink,
    UBIFS_ITYPE_BLK: File,
    UBIFS_ITYPE_CHR: File,
    UBIFS_ITYPE_FIFO: File,
    UBIFS_ITYPE_SOCK: File,
}



import re
from ctypes import BigEndianStructure, c_uint32

# Regex for a fdt_header with version=17 and last_comp_version=16.
RE_FDT_HEADER = re.compile(b"\xD0\x0D\xFE\xED.{16}\x00{3}\x11\x00{3}\x10.{12}", re.DOTALL)


class FDTHeader(BigEndianStructure):
    _fields_ = [
        ("magic", c_uint32),
        ("totalsize", c_uint32),
        ("off_dt_struct", c_uint32),
        ("off_dt_strings", c_uint32),
        ("off_mem_rsvmap", c_uint32),
        ("version", c_uint32),
        ("last_comp_version", c_uint32),
        ("boot_cpuid_phys", c_uint32),
        ("size_dt_strings", c_uint32),
        ("size_dt_struct", c_uint32),
    ]

    magic: int
    totalsize: int
    off_dt_struct: int
    off_dt_strings: int
    off_mem_rsvmap: int
    version: int
    last_comp_version: int
    boot_cpuid_phys: int
    size_dt_strings: int
    size_dt_struct: int



from __future__ import annotations

import hashlib
import io
from collections.abc import Callable, Generator
from contextlib import contextmanager
from enum import Enum
from os import scandir
from pathlib import Path
from shutil import disk_usage
from tempfile import gettempdir as _gettempdir
from typing import Any, AnyStr, BinaryIO, Optional, Union
from zipfile import is_zipfile

from lz4.block import decompress as lz4_block_decompress
from pakler import Section, is_pak_file
from pycramfs.const import MAGIC_BYTES as CRAMFS_MAGIC
from PySquashfsImage.const import SQUASHFS_MAGIC
from ubireader.ubi import ubi
from ubireader.ubi.defines import UBI_EC_HDR_MAGIC as UBI_MAGIC
from ubireader.ubi_io import ubi_file
from ubireader.ubifs.defines import UBIFS_NODE_MAGIC as UBIFS_MAGIC
from ubireader.utils import guess_peb_size

from reolinkfw.tmpfile import TempFile
from reolinkfw.typedefs import Buffer, GenericPath

ONEMIB = 1024**2
ONEGIB = 1024**3


class FileType(Enum):
    CRAMFS = CRAMFS_MAGIC
    LZ4_LEGACY_FRAME = b"\x02!L\x18"
    SQUASHFS = SQUASHFS_MAGIC.to_bytes(4, "little")
    UBI = UBI_MAGIC
    UBIFS = UBIFS_MAGIC
    UIMAGE = 0x27051956.to_bytes(4, "big")

    @classmethod
    def from_magic(cls, key: bytes, default: Optional[Any] = None) -> Optional[FileType]:
        try:
            return cls(key)
        except ValueError:
            return default


class SectionFile(io.BufferedIOBase):

    def __init__(self, fd: BinaryIO, section: Section, close: Callable[[BinaryIO], None]) -> None:
        self._fd = fd
        self._close = close
        self._start = section.start
        self._end = section.start + section.len
        self._position = section.start

    def peek(self, size: int = 0, /) -> bytes:
        if self._fd is None:
            raise ValueError("peek from closed file")
        if not isinstance(size, int):
            raise ValueError("size must be an int")
        if self._position >= self._end or size == 0:
            return b''
        max_read = self._end - self._position
        if size < 0 or size > max_read:
            size = max_read
        self._fd.seek(self._position)
        data = self._fd.read(size)
        return data

    def read(self, size: Optional[int] = -1, /) -> bytes:
        if self._fd is None:
            raise ValueError("read from closed file")
        if self._position >= self._end or size == 0:
            return b''
        max_read = self._end - self._position
        if size is None or size < 0 or size > max_read:
            size = max_read
        self._fd.seek(self._position)
        data = self._fd.read(size)
        self._position = self._fd.tell()
        return data

    def seek(self, offset: int, whence: int = io.SEEK_SET, /) -> int:
        if self._fd is None:
            raise ValueError("seek on closed file")
        if whence == io.SEEK_SET:
            newpos = self._start + offset
        elif whence == io.SEEK_CUR:
            newpos = self._position + offset
        elif whence == io.SEEK_END:
            newpos = self._end + offset
        else:
            raise ValueError(f"whence value {whence} unsupported")
        if newpos < self._start:
            raise ValueError("new position is negative")
        self._position = newpos
        return self.tell()

    def tell(self) -> int:
        if self._fd is None:
            raise ValueError("tell on closed file")
        return self._position - self._start

    def close(self) -> None:
        try:
            if self._fd is not None:
                self._close(self._fd)
                self._fd = None
        finally:
            super().close()


@contextmanager
def closing_ubifile(ubifile: ubi_file) -> Generator[ubi_file, Any, None]:
    try:
        yield ubifile
    finally:
        ubifile._fhandle.close()


def get_fs_from_ubi(fd: BinaryIO, size: int, offset: int = 0) -> bytes:
    """Return the first file system that sits on top of the UBI volume."""
    fd.seek(offset)
    binbytes = fd.read(size)
    with TempFile(binbytes) as t:
        block_size = guess_peb_size(t)
        with closing_ubifile(ubi_file(t, block_size)) as ubifile:
            ubi_obj = ubi(ubifile)
            volume = ubi_obj.images[0].volumes.popitem()[1]
            return b''.join(volume.reader(ubi_obj))


def dir_size(path: Union[GenericPath[AnyStr], int, None] = None) -> int:
    size = 0
    try:
        with scandir(path) as it:
            for entry in it:
                if entry.is_dir(follow_symlinks=False):
                    size += dir_size(entry.path)
                elif entry.is_file(follow_symlinks=False):
                    size += entry.stat().st_size
    except OSError:
        pass
    return size


def gettempdir() -> Path:
    return Path(_gettempdir()) / "reolinkfwcache"


def get_cache_file(url: str) -> Path:
    file = gettempdir() / hashlib.sha256(url.encode("utf8")).hexdigest()
    if is_zipfile(file) or is_pak_file(file):
        return file
    try:
        with open(file, 'r', encoding="utf8") as f:
            return gettempdir() / f.read(256)
    except (OSError, UnicodeDecodeError):
        return file


def has_cache(url: str) -> bool:
    return get_cache_file(url).is_file()


def make_cache_file(url: str, filebytes: Buffer, name: Optional[str] = None) -> bool:
    tempdir = gettempdir()
    tempdir.mkdir(exist_ok=True)
    if disk_usage(tempdir).free < ONEGIB or dir_size(tempdir) > ONEGIB:
        return False
    sha = hashlib.sha256(url.encode("utf8")).hexdigest()
    name = sha if not isinstance(name, str) else name
    try:
        with open(tempdir / name, "wb") as f:
            f.write(filebytes)
        if name != sha:
            with open(tempdir / sha, 'w', encoding="utf8") as f:
                f.write(name)
    except OSError:
        return False
    return True


def lz4_legacy_decompress(f: BinaryIO) -> bytes:
    # https://github.com/python-lz4/python-lz4/issues/169
    res = b''
    if f.read(4) != FileType.LZ4_LEGACY_FRAME.value:
        raise Exception("LZ4 legacy frame magic not found")
    while (size := int.from_bytes(f.read(4), "little")) != len(res):
        res += lz4_block_decompress(f.read(size), uncompressed_size=8*ONEMIB)
    return res



import collections.abc
from os import PathLike
from typing import AnyStr, Optional, TypedDict, Union

from aiohttp.typedefs import StrOrURL

if hasattr(collections.abc, "Buffer"):
    Buffer = collections.abc.Buffer
else:
    Buffer = Union[bytes, bytearray, memoryview]

GenericPath = Union[AnyStr, PathLike[AnyStr]]
StrPath = Union[str, PathLike[str]]
StrPathURL = Union[StrPath, StrOrURL]
StrOrBytesPath = Union[StrPath, bytes, PathLike[bytes]]
FileDescriptorOrPath = Union[int, StrOrBytesPath]

InfoFiles = TypedDict("InfoFiles", {
    "version_file": bytes,
    "dvr.xml": bytes,
    "dvr": Optional[bytes],
    "router": Optional[bytes],
})


class DVRInfo(TypedDict):
    """Info from the dvr.xml and version_file files."""
    board_name: str
    board_type: str
    build_date: str
    detail_machine_type: str
    display_type_info: str
    firmware_version_prefix: str
    type: str
    version_file: str



from __future__ import annotations

import os
import platform
import sys
import tempfile
from abc import abstractmethod

from reolinkfw.typedefs import Buffer


class TempFileFromBytes(os.PathLike):

    def __init__(self, filebytes: Buffer) -> None:
        self._filebytes = filebytes
        self._fd = -1
        self._path = None

    def __enter__(self) -> TempFileFromBytes:
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def __fspath__(self) -> str:
        return self._path

    @abstractmethod
    def open(self) -> int:
        ...

    @abstractmethod
    def close(self) -> None:
        ...


class LinuxInMemoryFile(TempFileFromBytes):

    def open(self) -> int:
        self._fd = os.memfd_create("temp")
        os.write(self._fd, self._filebytes)
        self._path = f"/proc/self/fd/{self._fd}"
        return self._fd

    def close(self) -> None:
        os.close(self._fd)
        self._fd = -1
        self._path = None


class OnDiskTempFile(TempFileFromBytes):

    def open(self) -> int:
        self._fd, self._path = tempfile.mkstemp()
        os.write(self._fd, self._filebytes)
        return self._fd

    def close(self) -> None:
        os.close(self._fd)
        os.unlink(self._path)
        self._fd = -1
        self._path = None


if sys.platform.startswith("linux") and platform.libc_ver()[1] >= "2.27":
    TempFile = LinuxInMemoryFile
else:
    TempFile = OnDiskTempFile



import asyncio
import gzip
import hashlib
import io
import lzma
import posixpath
import re
import zlib
from ast import literal_eval
from collections.abc import Iterator
from contextlib import redirect_stdout
from ctypes import sizeof
from functools import partial
from pathlib import Path
from typing import IO, Any, BinaryIO, Optional, Union
from urllib.parse import parse_qsl, urlparse
from zipfile import ZipFile, is_zipfile

import aiohttp
import pybcl
from aiohttp.typedefs import StrOrURL
from lxml.etree import fromstring
from lxml.html import document_fromstring
from pakler import PAK, Section, is_pak_file
from pycramfs import Cramfs
from pycramfs.extract import extract_dir as extract_cramfs
from pyfdt.pyfdt import Fdt, FdtBlobParse
from PySquashfsImage import SquashFsImage
from PySquashfsImage.extract import extract_dir as extract_squashfs
from ubireader.ubifs import ubifs as ubifs_
from ubireader.ubifs.output import extract_files as extract_ubifs
from ubireader.ubi_io import ubi_file
from ubireader.utils import guess_leb_size

from reolinkfw.fdt import RE_FDT_HEADER, FDTHeader
from reolinkfw.tmpfile import TempFile
from reolinkfw.typedefs import Buffer, DVRInfo, InfoFiles, StrPath, StrPathURL
from reolinkfw.ubifs import UBIFS
from reolinkfw.uboot import Compression, LegacyImageHeader, get_arch_name
from reolinkfw.util import (
    ONEMIB,
    FileType,
    SectionFile,
    closing_ubifile,
    get_cache_file,
    get_fs_from_ubi,
    has_cache,
    lz4_legacy_decompress,
    make_cache_file,
)

__version__ = "2.0.0"

FILES = ("version_file", "dvr.xml", "dvr", "router")
INFO_KEYS = ("firmware_version_prefix", "board_type", "board_name", "build_date", "display_type_info", "detail_machine_type", "type")

UBOOT_SECTIONS = ("uboot", "uboot1", "BOOT")
KERNEL_SECTIONS = ("kernel", "KERNEL")
ROOTFS_SECTIONS = ("fs", "rootfs")
FS_SECTIONS = ROOTFS_SECTIONS + ("app",)

RE_BANNER = re.compile(rb"\x00(Linux version .+? \(.+?@.+?\) \(.+?\) .+?)\n\x00")
RE_COMPLINK = re.compile(rb"\x00([^\x00]+?-linux-.+? \(.+?\) [0-9].+?)\n\x00+(.+?)\n\x00")
RE_IKCFG = re.compile(b"IKCFG_ST(.+?)IKCFG_ED", re.DOTALL)
RE_KERNEL_COMP = re.compile(
    b"(?P<lz4>" + FileType.LZ4_LEGACY_FRAME.value + b')'
    b"|(?P<xz>\xFD\x37\x7A\x58\x5A\x00\x00.(?!XZ))"
    b"|(?P<lzma>.{5}\xff{8})"
    b"|(?P<gzip>\x1f\x8b\x08\x00\x00\x00\x00\x00\x02\x03)"
)
RE_LZMA_OR_XZ = re.compile(b".{5}\xff{8}|\xFD\x37\x7A\x58\x5A\x00\x00")
# Pattern for a legacy image header with these properties:
# OS: U-Boot / firmware (0x11)
# Type: kernel (0x02)
# Only used for MStar/SigmaStar cameras (Lumus and RLC-410W IPC_30K128M4MP)
RE_MSTAR = re.compile(FileType.UIMAGE.value + b".{24}\x11.\x02.{33}", re.DOTALL)
RE_UBOOT = re.compile(rb"U-Boot [0-9]{4}\.[0-9]{2}.*? \(.+?\)")

DUMMY = object()


class ReolinkFirmware(PAK):

    def __init__(self, fd: BinaryIO, offset: int = 0, closefd: bool = True) -> None:
        super().__init__(fd, offset, closefd)
        self._uboot_section_name = self._get_uboot_section_name()
        self._uboot_section = None
        self._uboot = None
        self._kernel_section_name = self._get_kernel_section_name()
        self._kernel_section = None
        self._kernel = None
        self._fdt = DUMMY
        self._sdict = {s.name: s for s in self}
        self._open_files = 1
        self._fs_sections = [s for s in self if s.name in FS_SECTIONS]

    def __del__(self) -> None:
        self.close()

    def __getitem__(self, key: Union[int, str]) -> Section:
        if isinstance(key, int):
            return self.sections[key]
        if isinstance(key, str):
            if key.lower() == "uboot":
                key = self._uboot_section_name
            elif key.lower() == "kernel":
                key = self._kernel_section_name
            return self._sdict[key]
        raise TypeError

    def __iter__(self) -> Iterator[Section]:
        yield from self.sections

    @property
    def uboot_section(self) -> bytes:
        """Return the firmware's U-Boot section as bytes."""
        if self._uboot_section is not None:
            return self._uboot_section
        self._uboot_section = self.extract_section(self["uboot"])
        return self._uboot_section

    @property
    def uboot(self) -> bytes:
        """Return the firmware's decompressed U-Boot as bytes.

        If the U-Boot is not compressed this gives the same result
        as the `uboot_section` property.
        """
        if self._uboot is not None:
            return self._uboot
        self._uboot = self._decompress_uboot()
        return self._uboot

    @property
    def kernel_section(self) -> bytes:
        """Return the firmware's kernel section as bytes."""
        if self._kernel_section is not None:
            return self._kernel_section
        self._kernel_section = self.extract_section(self["kernel"])
        return self._kernel_section

    @property
    def kernel(self) -> bytes:
        """Return the firmware's decompressed kernel as bytes."""
        if self._kernel is not None:
            return self._kernel
        self._kernel = self._decompress_kernel()
        return self._kernel

    @property
    def fdt(self) -> Optional[Fdt]:
        if self._fdt is not DUMMY:
            return self._fdt  # type: ignore
        self._fdt = self._get_fdt()
        return self._fdt

    @property
    def fdt_json(self) -> Optional[dict[str, Any]]:
        if self.fdt is not None:
            return literal_eval(self.fdt.to_json().replace("null", "None"))
        return None

    @property
    def board(self) -> Optional[str]:
        if self.fdt_json is not None:
            return self.fdt_json["model"][1]
        return None

    def _fdclose(self, fd: BinaryIO) -> None:
        self._open_files -= 1
        if self._closefd and not self._open_files:
            fd.close()

    def _get_uboot_section_name(self) -> str:
        for section in self:
            if section.len and section.name in UBOOT_SECTIONS:
                return section.name
        raise Exception("U-Boot section not found")

    def _get_kernel_section_name(self) -> str:
        for section in self:
            if section.len and section.name in KERNEL_SECTIONS:
                return section.name
        raise Exception("Kernel section not found")

    def _decompress_uboot(self) -> bytes:
        uboot = self.uboot_section
        if uboot.startswith(pybcl.BCL_MAGIC_BYTES):
            # Sometimes section.len - sizeof(hdr) is 1 to 3 bytes larger
            # than hdr.size. The extra bytes are 0xff (padding?). This
            # could explain why the compressed size is added to the header.
            hdr = pybcl.HeaderVariant.from_buffer_copy(uboot)
            compressed = uboot[sizeof(hdr):sizeof(hdr)+hdr.size]
            return pybcl.decompress(compressed, hdr.algo, hdr.outsize)
        if (match := RE_MSTAR.search(uboot)) is not None:
            hdr = LegacyImageHeader.from_buffer_copy(uboot, match.start())
            start = match.start() + sizeof(hdr)
            if hdr.comp == Compression.LZMA:
                return lzma.decompress(uboot[start:start+hdr.size])
            raise Exception(f"Unexpected compression {hdr.comp}")
        return uboot  # Assume no compression

    def _decompress_kernel(self) -> bytes:
        # Use lzma.LZMADecompressor instead of lzma.decompress
        # because we know there's only one stream.
        data = self.kernel_section
        uimage_hdr_size = sizeof(LegacyImageHeader)
        # RLN36 kernel image headers report no compression
        # so don't bother reading the header and just look for
        # a compression magic.
        if RE_LZMA_OR_XZ.match(data, uimage_hdr_size):
            return lzma.LZMADecompressor().decompress(data[uimage_hdr_size:])
        if (halt := data.find(b" -- System halted")) == -1:
            raise Exception("'System halted' string not found")
        match = RE_KERNEL_COMP.search(data, halt)
        if match is None:
            raise Exception("No known compression found in kernel")
        start = match.start()
        if match.lastgroup == "lz4":
            return lz4_legacy_decompress(io.BytesIO(data[start:]))
        elif match.lastgroup in ("xz", "lzma"):
            return lzma.LZMADecompressor().decompress(data[start:])
        elif match.lastgroup == "gzip":
            # wbits=31 because only one member to decompress.
            return zlib.decompress(data[start:], wbits=31)
        raise Exception("unreachable")

    def _get_fdt(self) -> Optional[Fdt]:
        # At most 2 FDTs can be found in a firmware, and usually only one.
        # Most of the time it's in the fdt section or in the decompressed kernel.
        # Hardware versions starting with IPC_30 or IPC_32 have 1 FDT
        # in the decompressed kernel.
        # Hardware versions starting with IPC_35, IPC_36 or IPC_38 have no FDT.
        # HI3536CV100 -> only firmwares with FDT in kernel_section
        # Some firmwares with one FDT in the fdt section have a 2nd FDT
        # in the U-Boot section with no model.
        match = data = None
        if "fdt" in self._sdict:
            # Reolink Duo 1: 2 FDTs, section starts with header FKLR (Reolink FDT?)
            # Some NVRs: 2 FDTs
            data = self.extract_section(self["fdt"])
            match = RE_FDT_HEADER.search(data)
        elif (match := RE_FDT_HEADER.search(self.kernel_section)) is not None:
            data = self.kernel_section
        elif (match := RE_FDT_HEADER.search(self.kernel)) is not None:
            data = self.kernel
        if match is not None and data is not None:
            start = match.start()
            hdr = FDTHeader.from_buffer_copy(data, start)
            end = start + hdr.totalsize
            return FdtBlobParse(io.BytesIO(data[start:end])).to_fdt()
        return None

    def open(self, section: Union[Section, str]) -> SectionFile:
        if isinstance(section, str):
            section = self[section]
        self._open_files += 1
        return SectionFile(self._fd, section, self._fdclose)

    def close(self) -> None:
        if self._fd is not None:
            self._fdclose(self._fd)
            self._fd = None

    def sha256(self) -> str:
        sha = hashlib.sha256()
        self._fd.seek(0)
        for block in iter(partial(self._fd.read, ONEMIB), b''):
            sha.update(block)
        return sha.hexdigest()

    def get_uboot_info(self) -> tuple[Optional[str], Optional[str], Optional[str]]:
        # Should never be None.
        match_ub = RE_UBOOT.search(self.uboot)
        version = match_ub.group().decode() if match_ub is not None else None
        # Should only be None for HiSilicon devices.
        match_cl = RE_COMPLINK.search(self.uboot)
        compiler = match_cl.group(1).decode() if match_cl is not None else None
        linker = match_cl.group(2).decode() if match_cl is not None else None
        return version, compiler, linker

    def get_kernel_image_header(self) -> Optional[LegacyImageHeader]:
        with self.open("kernel") as f:
            data = f.read(sizeof(LegacyImageHeader))
        if FileType.from_magic(data[:4]) == FileType.UIMAGE:
            return LegacyImageHeader.from_buffer_copy(data)
        return None

    def get_kernel_image_header_info(self) -> tuple[Optional[str], Optional[str], Optional[str]]:
        hdr = self.get_kernel_image_header()
        if hdr is None:
            return None, None, None
        os = "Linux" if hdr.os == 5 else "Unknown"
        return os, get_arch_name(hdr.arch), hdr.name

    def get_linux_banner(self) -> Optional[str]:
        match = RE_BANNER.search(self.kernel)
        return match.group(1).decode() if match is not None else None

    def get_kernel_config(self) -> Optional[bytes]:
        if (match := RE_IKCFG.search(self.kernel)) is not None:
            return gzip.decompress(match.group(1))
        return None

    def get_vendor(self) -> Optional[str]:
        map_ = {
            "novatek": "Novatek",
            "sstar": "MStar/SigmaStar",
            "hisilicon": "HiSilicon",
        }
        if self.fdt_json is not None:
            key = self.fdt_json["compatible"][1].split(',')[0]
            return map_.get(key.lower(), key)
        with self.open("uboot") as f:
            if re.match(b"GM[0-9]{4}", f.read(6)):
                return "Grain Media"
        if re.search(b"HISILICON LOGO MAGIC", self.uboot_section) is not None:
            return "HiSilicon"
        return None

    def get_fs_info(self) -> list[dict[str, str]]:
        result = []
        for section in self._fs_sections:
            with self.open(section) as f:
                fs = FileType.from_magic(f.read(4))
                if fs == FileType.UBI:
                    f.seek(266240)
                    fs = FileType.from_magic(f.read(4))
            result.append({
                "name": section.name,
                "type": fs.name.lower() if fs is not None else "unknown"
            })
        return result

    async def get_info(self) -> dict[str, Any]:
        ha = await asyncio.to_thread(self.sha256)
        app = self._fs_sections[-1]
        with self.open(app) as f:
            fs = FileType.from_magic(f.read(4))
            if fs == FileType.CRAMFS:
                files = await asyncio.to_thread(get_files_from_cramfs, f, 0, False)
            elif fs == FileType.UBI:
                files = await asyncio.to_thread(get_files_from_ubi, f, app.len, 0)
            elif fs == FileType.SQUASHFS:
                files = await asyncio.to_thread(get_files_from_squashfs, f, 0, False)
            else:
                return {"error": "Unrecognized image type", "sha256": ha}
        os, architecture, kernel_image_name = self.get_kernel_image_header_info()
        uboot_version, compiler, linker = self.get_uboot_info()
        return {
            **get_info_from_files(files),
            "os": os,
            "architecture": architecture,
            "kernel_image_name": kernel_image_name,
            "uboot_version": uboot_version,
            "uboot_compiler": compiler,
            "uboot_linker": linker,
            "linux_banner": self.get_linux_banner(),
            "board": self.board,
            "board_vendor": self.get_vendor(),
            "filesystems": self.get_fs_info(),
            "sha256": ha
        }

    def extract_file_system(self, section: Section, dest: Optional[Path] = None) -> None:
        dest = (Path.cwd() / "reolink_fs") if dest is None else dest
        dest.mkdir(parents=True, exist_ok=True)
        with self.open(section) as f:
            fs = FileType.from_magic(f.read(4))
            if fs == FileType.UBI:
                fs_bytes = get_fs_from_ubi(f, section.len, 0)
                fs = FileType.from_magic(fs_bytes[:4])
                if fs == FileType.UBIFS:
                    with TempFile(fs_bytes) as file:
                        block_size = guess_leb_size(file)
                        with closing_ubifile(ubi_file(file, block_size)) as ubifile:
                            with redirect_stdout(io.StringIO()):
                                # Files that already exist are not written again.
                                extract_ubifs(ubifs_(ubifile), dest)
                elif fs == FileType.SQUASHFS:
                    with SquashFsImage.from_bytes(fs_bytes) as image:
                        extract_squashfs(image.root, dest, True)
                else:
                    raise Exception("Unknown file system in UBI")
            elif fs == FileType.SQUASHFS:
                with SquashFsImage(f, 0, False) as image:
                    extract_squashfs(image.root, dest, True)
            elif fs == FileType.CRAMFS:
                with Cramfs.from_fd(f, 0, False) as image:
                    extract_cramfs(image.rootdir, dest, True)
            else:
                raise Exception("Unknown file system")

    def extract(self, dest: Optional[Path] = None, force: bool = False) -> None:
        dest = (Path.cwd() / "reolink_firmware") if dest is None else dest
        dest.mkdir(parents=True, exist_ok=force)
        rootfsdir = [s.name for s in self if s.name in ROOTFS_SECTIONS][0]
        for section in self._fs_sections:
            if section.name == "app":
                outpath = dest / rootfsdir / "mnt" / "app"
            else:
                outpath = dest / rootfsdir
            self.extract_file_system(section, outpath)
        mode = "wb" if force else "xb"
        with open(dest / "uboot", mode) as f:
            f.write(self.uboot)
        with open(dest / "kernel", mode) as f:
            f.write(self.kernel)
        if (kcfg := self.get_kernel_config()) is not None:
            with open(dest / ".config", mode) as f:
                f.write(kcfg)
        if self.fdt is not None:
            with open(dest / "camera.dts", 'w' if force else 'x', encoding="utf8") as f:
                f.write(self.fdt.to_dts())


async def download(url: StrOrURL) -> Union[bytes, int]:
    """Return resource as bytes.

    Return the status code of the request if it is not 200.
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as resp:
            return await resp.read() if resp.status == 200 else resp.status


def firmwares_from_zip(zip: Union[StrPath, IO[bytes]]) -> list[tuple[str, ReolinkFirmware]]:
    """Return a list of tuples, one for each firmware found in the ZIP.

    It is the caller's responsibility to close the firmware files.
    """
    fws = []
    with ZipFile(zip) as myzip:
        for name in myzip.namelist():
            file = myzip.open(name)
            if is_pak_file(file):
                fws.append((file.name, ReolinkFirmware.from_fd(file)))
            else:
                file.close()
    return fws


def get_info_from_files(files: InfoFiles) -> DVRInfo:
    xml: dict[str, str] = dict(fromstring(files["dvr.xml"]).items())
    info = {k: xml.get(k) for k in INFO_KEYS}
    info["version_file"] = files["version_file"].decode().strip()
    if not info.get("firmware_version_prefix"):
        thefile = files["dvr"] if files["dvr"] is not None else files["router"]
        match = re.search(rb"echo (v[23]\.0\.0)", thefile) if thefile is not None else None
        info["firmware_version_prefix"] = match.group(1).decode() if match else None
    return info  # type: ignore


def get_files_from_squashfs(fd: BinaryIO, offset: int = 0, closefd: bool = True) -> InfoFiles:
    # Firmwares using squashfs have either one or two file system
    # sections. When there is only one, the app directory is located at
    # /mnt/app. Otherwise it's the same as with cramfs and ubifs.
    files = dict.fromkeys(FILES)
    with SquashFsImage(fd, offset, closefd) as image:
        for name in files:
            path2 = posixpath.join("/mnt/app", name)
            if (file := (image.select(name) or image.select(path2))) is not None:
                files[name] = file.read_bytes()
    return files  # type: ignore


def get_files_from_ubifs(binbytes: Buffer) -> InfoFiles:
    # For now all firmwares using ubifs have two file system sections.
    # The interesting files are in the root directory of the "app" one.
    # Using select() with a relative path is enough.
    files = dict.fromkeys(FILES)
    with TempFile(binbytes) as tempfile:
        with UBIFS.from_file(tempfile) as image:
            for name in files:
                if (file := image.select(name)) is not None:
                    files[name] = file.read_bytes()
    return files  # type: ignore


def get_files_from_ubi(fd: BinaryIO, size: int, offset: int = 0) -> InfoFiles:
    fsbytes = get_fs_from_ubi(fd, size, offset)
    fs = FileType.from_magic(fsbytes[:4])
    if fs == FileType.UBIFS:
        return get_files_from_ubifs(fsbytes)
    elif fs == FileType.SQUASHFS:
        return get_files_from_squashfs(io.BytesIO(fsbytes))
    raise Exception("Unknown file system in UBI")


def get_files_from_cramfs(fd: BinaryIO, offset: int = 0, closefd: bool = True) -> InfoFiles:
    # For now all firmwares using cramfs have two file system sections.
    # The interesting files are in the root directory of the "app" one.
    # Using select() with a relative path is enough.
    files = dict.fromkeys(FILES)
    with Cramfs.from_fd(fd, offset, closefd) as cramfs:
        for name in files:
            if (file := cramfs.select(name)) is not None:
                files[name] = file.read_bytes()
    return files  # type: ignore


def is_url(string: StrOrURL) -> bool:
    return str(string).startswith("http")


def is_local_file(string: StrPath) -> bool:
    return Path(string).is_file()


async def direct_download_url(url: str) -> str:
    if url.startswith("https://drive.google.com/file/d/"):
        return f"https://drive.google.com/uc?id={url.split('/')[5]}&confirm=t"
    elif url.startswith("https://www.mediafire.com/file/"):
        doc = document_fromstring(await download(url))
        return doc.get_element_by_id("downloadButton").get("href")
    elif url.startswith("https://bit.ly/"):
        async with aiohttp.ClientSession() as session:
            async with session.get(url, allow_redirects=False) as resp:
                return await direct_download_url(resp.headers["Location"])
    return url


async def firmwares_from_file(file_or_url: StrPathURL, use_cache: bool = True) -> list[tuple[Optional[str], ReolinkFirmware]]:
    """Return firmwares read from an on-disk file or a URL.

    The file or resource may be a ZIP or a PAK. On success return a
    list of 2-tuples where each tuple is of the form
    `(filename, firmware)`. When the argument is a URL, `filename` may
    be None. If the file is a ZIP the list might be empty.
    It is the caller's responsibility to close the firmware files.
    """
    file_or_url = str(file_or_url)
    if is_url(file_or_url):
        if use_cache and has_cache(file_or_url):
            return await firmwares_from_file(get_cache_file(file_or_url))
        file_or_url = await direct_download_url(file_or_url)
        zip_or_pak_bytes = await download(file_or_url)
        if isinstance(zip_or_pak_bytes, int):
            raise Exception(f"HTTP error {zip_or_pak_bytes}")
        pakname = dict(parse_qsl(urlparse(file_or_url).query)).get("name")
        if use_cache:
            make_cache_file(file_or_url, zip_or_pak_bytes, pakname)
        if is_pak_file(zip_or_pak_bytes):
            return [(pakname, ReolinkFirmware.from_bytes(zip_or_pak_bytes))]
        else:
            zipfile = io.BytesIO(zip_or_pak_bytes)
            if is_zipfile(zipfile):
                return await asyncio.to_thread(firmwares_from_zip, zipfile)
            zipfile.close()
            raise Exception("Not a ZIP or a PAK file")
    elif is_local_file(file_or_url):
        file_or_url = Path(file_or_url)
        if is_zipfile(file_or_url):
            return await asyncio.to_thread(firmwares_from_zip, file_or_url)
        elif is_pak_file(file_or_url):
            return [(file_or_url.name, ReolinkFirmware.from_file(file_or_url))]
        raise Exception("Not a ZIP or a PAK file")
    raise Exception("Not a URL or file")


async def firmware_info(file_or_url: StrPathURL, use_cache: bool = True) -> list[dict[str, Any]]:
    """Retrieve firmware info from an on-disk file or a URL.

    The file or resource may be a ZIP or a PAK.
    """
    try:
        fws = await firmwares_from_file(file_or_url, use_cache)
    except Exception as e:
        return [{"file": file_or_url, "error": str(e)}]
    if not fws:
        return [{"file": file_or_url, "error": "No PAKs found in ZIP file"}]
    info = [{**await fw.get_info(), "file": file_or_url, "pak": pakname} for pakname, fw in fws]
    for _, fw in fws:
        fw.close()
    return info



from __future__ import annotations

from ctypes import BigEndianStructure, c_char, c_uint32, c_uint8, sizeof
from enum import IntEnum
from typing import BinaryIO

UBOOT_MAGIC = 0x27051956


class Arch(IntEnum):
    ARM = 2
    MIPS = 5
    ARM64 = 22


class Compression(IntEnum):
    LZMA = 3


class LegacyImageHeader(BigEndianStructure):
    _fields_ = [
        ("_magic", c_uint32),
        ("_hcrc", c_uint32),
        ("_time", c_uint32),
        ("_size", c_uint32),
        ("_load", c_uint32),
        ("_ep", c_uint32),
        ("_dcrc", c_uint32),
        ("_os", c_uint8),
        ("_arch", c_uint8),
        ("_type", c_uint8),
        ("_comp", c_uint8),
        ("_name", c_char * 32),
    ]

    _magic: int
    _hcrc: int
    _time: int
    _size: int
    _load: int
    _ep: int
    _dcrc: int
    _os: int
    _arch: int
    _type: int
    _comp: int
    _name: bytes

    @property
    def magic(self) -> int:
        return self._magic

    @property
    def hcrc(self) -> int:
        return self._hcrc

    @property
    def time(self) -> int:
        return self._time

    @property
    def size(self) -> int:
        return self._size

    @property
    def load(self) -> int:
        return self._load

    @property
    def ep(self) -> int:
        return self._ep

    @property
    def dcrc(self) -> int:
        return self._dcrc

    @property
    def os(self) -> int:
        return self._os

    @property
    def arch(self) -> Arch:
        return Arch(self._arch)

    @property
    def type(self) -> int:
        return self._type

    @property
    def comp(self) -> int:
        return self._comp

    @property
    def name(self) -> str:
        return self._name.decode()

    @classmethod
    def from_fd(cls, fd: BinaryIO) -> LegacyImageHeader:
        return cls.from_buffer_copy(fd.read(sizeof(cls)))


def get_arch_name(arch: Arch) -> str:
    if arch == Arch.ARM:
        return "ARM"
    elif arch == Arch.MIPS:
        return "MIPS"
    elif arch == Arch.ARM64:
        return "AArch64"
    raise Exception("Unknown architecture")
